from .context import GluePySparkContext
from .job import GluePySparkJob
from .mixins import (
    CSVFormatOptions,
    JDBCConnectionOptions,
    JSONFormatOptions,
    ParquetFormatOptions,
    S3FormatOptions,
    S3SinkConnectionOptions,
    S3SourceConnectionOptions,
    XMLFormatOptions,
)

__all__ = [
    "GluePySparkContext",
    "GluePySparkJob",
    "JDBCConnectionOptions",
    "S3SourceConnectionOptions",
    "S3SinkConnectionOptions",
    "S3FormatOptions",
    "CSVFormatOptions",
    "JSONFormatOptions",
    "ParquetFormatOptions",
    "XMLFormatOptions",
]



from enum import Enum  # noqa: D100


class ConnectionType(Enum):
    """Enum class representing different types of connections."""

    DYNAMODB = "dynamodb"
    KINESIS = "kinesis"
    S3 = "s3"
    DOCUMENTDB = "documentdb"
    OPENSEARCH = "opensearch"
    REDSHIFT = "redshift"
    KAFKA = "kafka"
    AZURE_COSMOS = "azurecosmos"
    AZURE_SQL = "azuresql"
    BIGQUERY = "bigquery"
    MONGODB = "mongodb"
    SQLSERVER = "sqlserver"
    MYSQL = "mysql"
    ORACLE = "oracle"
    POSTGRESQL = "postgresql"
    SAPHANA = "saphana"
    SNOWFLAKE = "snowflake"
    TERADATA = "teradata"
    VERTICA = "vertica"
    MARKETPLACE_ATHENA = "marketplace.athena"
    MARKETPLACE_SPARK = "marketplace.spark"
    MARKETPLACE_JDBC = "marketplace.jdbc"
    CUSTOM_ATHENA = "custom.athena"
    CUSTOM_SPARK = "custom.spark"
    CUSTOM_JDBC = "custom.jdbc"



"""Module providing the GluePySparkJob class for handling Glue ETL jobs."""

from __future__ import annotations

import sys
from contextlib import contextmanager
from dataclasses import fields
from typing import TYPE_CHECKING, Generic, cast, overload

from awsglue.job import Job
from awsglue.utils import getResolvedOptions
from pyspark import SparkContext
from typing_extensions import TypeVar

from glue_utils import BaseOptions
from glue_utils.pyspark.context import GluePySparkContext

if TYPE_CHECKING:
    from collections.abc import Generator

    from pyspark.sql import SparkSession

T = TypeVar("T", bound=BaseOptions, default=BaseOptions)


class GluePySparkJob(Generic[T]):
    """Class that handles the boilerplate setup for Glue ETL jobs."""

    options: T
    sc: SparkContext
    spark: SparkSession
    glue_context: GluePySparkContext

    @overload
    def __init__(
        self: GluePySparkJob[BaseOptions],
    ) -> None: ...

    @overload
    def __init__(
        self: GluePySparkJob[T],
        *,
        options_cls: type[T],
    ) -> None: ...

    def __init__(
        self,
        *,
        options_cls: type[T | BaseOptions] = BaseOptions,
    ) -> None:
        """Initialize the GlueETLJob.

        Parameters
        ----------
        options_cls, optional
            Has to be a subclass of BaseOptions, by default BaseOptions

        """
        if not issubclass(options_cls, BaseOptions):
            msg = "options_cls must be a subclass of BaseOptions."
            raise TypeError(msg)

        field_names = {field.name for field in fields(options_cls)}

        params = []
        if "--JOB_NAME" in sys.argv:
            params.append("JOB_NAME")
        params.extend(field_names)

        _options = getResolvedOptions(sys.argv, params)

        self.options = cast(T, options_cls.from_options(_options))

        self.sc = SparkContext.getOrCreate()
        self.glue_context = GluePySparkContext(self.sc)
        self.spark = self.glue_context.spark_session

        self._job = Job(self.glue_context)
        self._job.init(_options.get("JOB_NAME", ""), _options)

    @contextmanager
    def managed_glue_context(
        self,
        *,
        commit: bool = True,
    ) -> Generator[GluePySparkContext, None, None]:
        """Context manager for managing the GluePySparkContext.

        Parameters
        ----------
        commit, optional
            Whether to commit the job, by default True

        """
        yield self.glue_context
        if commit:
            self.commit()

    def commit(self) -> None:
        """Commit the Glue ETL job."""
        self._job.commit()



from __future__ import annotations  # noqa: D100

from typing import TYPE_CHECKING, Literal, TypedDict

from .connection_types import ConnectionType

if TYPE_CHECKING:
    from awsglue.context import GlueContext
    from awsglue.dynamicframe import DynamicFrame


class JDBCConnectionOptions(TypedDict, total=False):
    """Connection options for JDBC connections.

    Parameters
    ----------
    useConnectionProperties : Literal["true"], optional
        Whether to use connection properties. Defaults to None.
    connectionName : str, optional
        The name of the connection. Defaults to None.
    url : str, optional
        The JDBC URL. Defaults to None.
    dbtable : str, optional
        The name of the table. Defaults to None.
    user : str, optional
        The username for the connection. Defaults to None.
    password : str, optional
        The password for the connection. Defaults to None.
    customJdbcDriverS3Path : str, optional
        The S3 path for the custom JDBC driver. Defaults to None.
    customJdbcDriverClassName : str, optional
        The class name for the custom JDBC driver. Defaults to None.
    bulkSize : int, optional
        The bulk size for the connection. Defaults to None.
    hashfield : str, optional
        The hash field for the connection. Defaults to None.
    hashexpression : str, optional
        The hash expression for the connection. Defaults to None.
    hashpartitions : int, optional
        The number of hash partitions for the connection. Defaults to None.
    sampleQuery : str, optional
        The sample query for the connection. Defaults to None.
    enablePartitioningForSampleQuery : bool, optional
        Whether to enable partitioning for the sample query. Defaults to None.
    sampleSize : int, optional
        The sample size for the connection. Defaults to None.

    Reference: https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-jdbc-home.html

    """

    useConnectionProperties: Literal["true"]
    connectionName: str
    url: str
    dbtable: str
    user: str
    password: str
    customJdbcDriverS3Path: str
    customJdbcDriverClassName: str
    bulkSize: int
    hashfield: str
    hashexpression: str
    hashpartitions: int
    sampleQuery: str
    enablePartitioningForSampleQuery: bool
    sampleSize: int
    jobBookmarkKeys: list[str]
    jobBookmarkKeysSortOrder: Literal["asc", "desc"]


class SQLServerMixin:
    """Mixin for working with SQL Server connections."""

    def create_dynamic_frame_from_sqlserver(
        self: GlueContext,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from a SQL Server database connection.

        Parameters
        ----------
        connection_options : JDBCConnectionOptions
            The connection options for the SQL Server database.
        transformation_ctx : str, optional
            The name of the transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the SQL Server database.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.SQLSERVER.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_sqlserver(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to a SQL Server database using JDBC.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write to the SQL Server database.
        connection_options : JDBCConnectionOptions
            The JDBC connection options for the SQL Server database.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame that was written to the SQL Server database.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.SQLSERVER.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )


class MySQLMixin:
    """Mixin for working with MySQL connections."""

    def create_dynamic_frame_from_mysql(
        self: GlueContext,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from a MySQL database connection.

        Parameters
        ----------
        connection_options : JDBCConnectionOptions
            The connection options for the MySQL database.
        transformation_ctx : str, optional
            The transformation context for the DynamicFrame. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the MySQL database.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.MYSQL.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_mysql(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to a MySQL database using JDBC.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write to the MySQL database.
        connection_options : JDBCConnectionOptions
            The JDBC connection options for the MySQL database.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame after writing to the MySQL database.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.MYSQL.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )


class OracleMixin:
    """Mixin for working with Oracle connections."""

    def create_dynamic_frame_from_oracle(
        self: GlueContext,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from an Oracle database connection.

        Parameters
        ----------
        connection_options : JDBCConnectionOptions
            The connection options for connecting to the Oracle database.
        transformation_ctx : str, optional
            The name of the transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the Oracle database.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.ORACLE.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_oracle(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to an Oracle database using JDBC.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write to the Oracle database.
        connection_options : JDBCConnectionOptions
            The JDBC connection options for connecting to the Oracle database.
        transformation_ctx : str, optional
            The transformation context for the write operation. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame that was written to the Oracle database.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.ORACLE.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )


class PostgreSQLMixin:
    """Mixin for working with PostgreSQL connections."""

    def create_dynamic_frame_from_postgresql(
        self: GlueContext,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from a PostgreSQL database connection.

        Parameters
        ----------
        connection_options : JDBCConnectionOptions
            The connection options for the PostgreSQL database.
        transformation_ctx : str, optional
            The transformation context for the DynamicFrame. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the PostgreSQL database.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.POSTGRESQL.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_postgresql(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: JDBCConnectionOptions,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to a PostgreSQL database.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write to the database.
        connection_options : JDBCConnectionOptions
            The JDBC connection options for the database.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame that was written to the database.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.POSTGRESQL.value,
            connection_options=connection_options,
            transformation_ctx=transformation_ctx,
        )


class JDBCMixin(MySQLMixin, OracleMixin, PostgreSQLMixin, SQLServerMixin):
    """Mixin for working with JDBC connections."""



"""Module containing GluePySparkContext class that extends GlueContext."""

from __future__ import annotations

from awsglue.context import GlueContext

from .mixins.jdbc import JDBCMixin
from .mixins.s3 import S3Mixin


class GluePySparkContext(
    JDBCMixin,
    S3Mixin,
    # error: Class cannot subclass "AWSGlueContext" (has type "Any")  [misc]
    GlueContext,  # type: ignore[misc]
):
    """A custom PySpark context that extends the functionality of the AWS GlueContext.

    This class combines the features of the mixin classes with the AWS
    GlueContext to provide additional functionality for working with
    JDBC and S3 data sources in a Glue job.
    """



"""Module for conveniently parsing options resolved from command-line arguments."""

from __future__ import annotations

import sys
from dataclasses import dataclass, fields
from typing import Any

from awsglue.utils import getResolvedOptions
from typing_extensions import Self


@dataclass
class BaseOptions:
    """Dataclass for storing resolved options."""

    @classmethod
    def from_sys_argv(cls) -> Self:
        """Create an instance of the class from Glue's resolved arguments."""
        resolved_options = getResolvedOptions(
            sys.argv, [field.name for field in fields(cls)]
        )

        return cls.from_options(resolved_options)

    @classmethod
    def from_options(cls, options: dict[str, Any] | None = None) -> Self:
        """Create an instance of the class from the provided options."""
        if not options:
            return cls()

        field_names = {field.name for field in fields(cls)}

        return cls(
            **{key: value for key, value in options.items() if key in field_names},
        )



from __future__ import annotations  # noqa: D100

from enum import Enum
from typing import TYPE_CHECKING, Literal, TypedDict

from .connection_types import ConnectionType

if TYPE_CHECKING:
    from awsglue.context import GlueContext
    from awsglue.dynamicframe import DynamicFrame


class Format(Enum):
    """Enum representing different file formats."""

    AVRO = "avro"
    CSV = "csv"
    JSON = "json"
    PARQUET = "parquet"
    ORC = "orc"
    XML = "xml"
    GROKLOG = "grokLog"
    ION = "ion"


class S3SourceConnectionOptions(TypedDict, total=False):
    """Connection options for S3 connections.

    Parameters
    ----------
    paths : list[str]
        List of S3 paths to include.
    exclusions : str
        S3 path exclusion pattern.
    compressionType : Literal["gzip", "bzip2"]
        Compression type for the files.
    groupFiles : Literal["inPartition", "none"]
        Grouping strategy for the files.
    groupSize : str
        Size of each group.
    recurse : bool
        Whether to recursively search for files in subdirectories.
    maxBand : int
        Maximum number of bands to use.
    maxFilesInBand : int
        Maximum number of files in each band.
    isFailFast : bool
        Whether to fail fast on errors.
    catalogPartitionPredicate : str
        Predicate for filtering partitions.
    excludeStorageClasses : list[str]
        List of storage classes to exclude.

    References
    ----------
    - AWS Glue Programming ETL Connect to S3:
      https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-s3-home.html

    """

    paths: list[str]
    exclusions: str
    compressionType: Literal["gzip", "bzip2"]
    groupFiles: Literal["inPartition", "none"]
    groupSize: str
    recurse: bool
    maxBand: int
    maxFilesInBand: int
    isFailFast: bool
    catalogPartitionPredicate: str
    excludeStorageClasses: list[str]


class S3SinkConnectionOptions(TypedDict, total=False):
    """Connection options for writing data to an S3 sink.

    Parameters
    ----------
    path : str
        The S3 path where the data will be written.
    compression : {'gzip', 'bzip2'}, optional
        The compression type to be used for the data.
    partitionKeys : list[str], optional
        The list of partition keys for the data.

    """

    path: str
    compression: Literal["gzip", "bzip2"]
    partitionKeys: list[str]


class S3FormatOptions(TypedDict, total=False):
    """Options for configuring the base format.

    Parameters
    ----------
    attachFilename : str, optional
        Specifies whether to attach the filename to the data. Defaults to False.
    attachTimestamp : str, optional
        Specifies whether to attach the timestamp to the data. Defaults to False.

    """

    attachFilename: str
    attachTimestamp: str


class CSVFormatOptions(S3FormatOptions, total=False):
    """Format options for CSV files.

    Parameters
    ----------
    separator : str, optional
        The separator character used in the CSV file. Defaults to ','.
    escaper : str, optional
        The character used to escape special characters in the CSV file. Defaults to '"'.
    quoteChar : str, optional
        The character used to quote fields in the CSV file. Defaults to '"'.
    multiLine : bool, optional
        Whether the CSV file can contain multiline records. Defaults to False.
    withHeader : bool, optional
        Whether the CSV file has a header row. Defaults to False.
    writeHeader : bool, optional
        Whether to write the header row when writing data to the CSV file. Defaults to False.
    skipFirst : bool, optional
        Whether to skip the first line of the CSV file. Defaults to False.
    optimizePerformance : bool, optional
        Whether to optimize performance when reading the CSV file. Defaults to False.
    strictCheckForQuoting : bool, optional
        Whether to perform strict checking for quoting in the CSV file. Defaults to False.

    Reference
    ---------
    - AWS Glue Programming ETL Connect to CSV:
      https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-csv-home.html

    """

    separator: str
    escaper: str
    quoteChar: str
    multiLine: bool
    withHeader: bool
    writeHeader: bool
    skipFirst: bool
    optimizePerformance: bool
    strictCheckForQuoting: bool


class S3CSVMixin:
    """Mixin for working with CSV files in S3."""

    def create_dynamic_frame_from_s3_csv(
        self: GlueContext,
        connection_options: S3SourceConnectionOptions,
        format_options: CSVFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from a CSV file stored in Amazon S3.

        Parameters
        ----------
        connection_options : S3SourceConnectionOptions
            The connection options for the S3 source.
        format_options : CSVFormatOptions | None, optional
            The format options for the CSV file. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the CSV file in S3.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.CSV.value,
            format_options=format_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_s3_csv(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: S3SourceConnectionOptions,
        format_options: CSVFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to a CSV file in Amazon S3.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write.
        connection_options : S3SourceConnectionOptions
            The connection options for the S3 source.
        format_options : CSVFormatOptions | None, optional
            The format options for the CSV file. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The written DynamicFrame.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.CSV.value,
            format_options=format_options,
            transformation_ctx=transformation_ctx,
        )


class ParquetFormatOptions(S3FormatOptions, total=False):
    """Format options for Parquet files.

    Parameters
    ----------
    useGlueParquetWriter : bool, optional
        Specifies whether to use the Glue Parquet writer. Defaults to False.
    compression : {'uncompressed', 'snappy', 'gzip', 'lzo'}, optional
        The compression type for the Parquet file. Defaults to 'uncompressed'.
    blockSize : int, optional
        The size of the Parquet file block. Defaults to 128 MB.
    pageSize : int, optional
        The size of the Parquet file page. Defaults to 1 MB.

    Reference
    ---------
    - AWS Glue Programming ETL Connect to Parquet:
      https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-parquet-home.html

    """

    useGlueParquetWriter: bool
    compression: Literal[
        "uncompressed",
        "snappy",
        "gzip",
        "lzo",
    ]
    blockSize: int
    pageSize: int


class S3ParquetMixin:
    """Mixin for working with Parquet files in S3."""

    def create_dynamic_frame_from_s3_parquet(
        self: GlueContext,
        connection_options: S3SourceConnectionOptions,
        format_options: ParquetFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from Parquet files stored in Amazon S3.

        Parameters
        ----------
        connection_options : S3SourceConnectionOptions
            The connection options for the S3 source.
        format_options : ParquetFormatOptions | None, optional
            The format options for reading Parquet files. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the Parquet files in S3.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.PARQUET.value,
            format_options=format_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_s3_parquet(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: S3SinkConnectionOptions,
        format_options: ParquetFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to Parquet format in Amazon S3.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to write.
        connection_options : S3SinkConnectionOptions
            The connection options for writing to S3.
        format_options : ParquetFormatOptions | None, optional
            The format options for writing Parquet. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The written DynamicFrame.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.PARQUET.value,
            format_options=format_options,
            transformation_ctx=transformation_ctx,
        )


class JSONFormatOptions(S3FormatOptions, total=False):
    """Format options for JSON files.

    Parameters
    ----------
    jsonPath : str, optional
        The JSON path to extract data from. Defaults to None.
    multiline : bool, optional
        Whether the JSON file can contain multiline records. Defaults to False.
    optimizePerformance : bool, optional
        Whether to optimize performance when reading the JSON file. Defaults to False.
    withSchema : str, optional
        The schema to use for reading the JSON file. Defaults to None.

    Reference
    ---------
    - AWS Glue Programming ETL Connect to JSON:
      https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-connect-json-home.html

    """

    jsonPath: str
    multiline: bool
    optimizePerformance: bool
    withSchema: str


class S3JSONMixin:
    """Mixin for working with JSON files in S3."""

    def create_dynamic_frame_from_s3_json(
        self: GlueContext,
        connection_options: S3SourceConnectionOptions,
        format_options: JSONFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from JSON data stored in Amazon S3.

        Parameters
        ----------
        connection_options : S3SourceConnectionOptions
            The connection options for the S3 source.
        format_options : JSONFormatOptions | None, optional
            The format options for the JSON data. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame created from the JSON data in S3.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.JSON.value,
            format_options=format_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_s3_json(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: S3SinkConnectionOptions,
        format_options: JSONFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to JSON format and save it in S3.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to be written.
        connection_options : S3SinkConnectionOptions
            The connection options for S3 sink.
        format_options : JSONFormatOptions | None, optional
            The format options for JSON. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame after writing to S3.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.JSON.value,
            format_options=format_options,
            transformation_ctx=transformation_ctx,
        )


class XMLFormatOptions(S3FormatOptions, total=False):
    """Format options for XML files.

    Parameters
    ----------
    rowTag : str, optional
        The XML tag that represents each row of data. Defaults to None.
    encoding : str, optional
        The encoding of the XML file. Defaults to None.
    excludeAttribute : bool, optional
        Whether to exclude XML attributes from the resulting DynamicFrame. Defaults to False.
    treatEmptyValueAsNull : bool, optional
        Whether to treat empty XML values as null values. Defaults to False.
    attributePrefix : str, optional
        The prefix to add to XML attributes when converting to DynamicFrame. Defaults to None.
    valueTag : str, optional
        The XML tag that represents the value of each field. Defaults to None.
    ignoreSurroundingSpaces : bool, optional
        Whether to ignore surrounding spaces when parsing XML values. Defaults to False.
    withSchema : str, optional
        The schema to use for reading the XML file. Defaults to None.

    Reference
    ---------
    - AWS Glue Programming ETL Connect to XML:
      https://docs.aws.amazon.com/glue/latest/dg/aws-glue-programming-etl-format-xml-home.html#aws-glue-programming-etl-format-xml-reference

    """

    rowTag: str
    encoding: str
    excludeAttribute: bool
    treatEmptyValueAsNull: bool
    attributePrefix: str
    valueTag: str
    ignoreSurroundingSpaces: bool
    withSchema: str


class S3XMLMixin:
    """Mixin for working with XML files in S3."""

    def create_dynamic_frame_from_s3_xml(
        self: GlueContext,
        connection_options: S3SourceConnectionOptions,
        format_options: XMLFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Create a DynamicFrame from XML data stored in Amazon S3.

        Parameters
        ----------
        connection_options : S3SourceConnectionOptions
            The connection options for the S3 data source.
        format_options : XMLFormatOptions, optional
            The format options for parsing the XML data. Defaults to None.
        transformation_ctx : str, optional
            The transformation context for the DynamicFrame. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame containing the parsed XML data.

        """
        return self.create_dynamic_frame_from_options(
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.XML.value,
            format_options=format_options,
            transformation_ctx=transformation_ctx,
        )

    def write_dynamic_frame_to_s3_xml(
        self: GlueContext,
        frame: DynamicFrame,
        connection_options: S3SinkConnectionOptions,
        format_options: XMLFormatOptions | None = None,
        transformation_ctx: str = "",
    ) -> DynamicFrame:
        """Write a DynamicFrame to XML format and saves it in S3.

        Parameters
        ----------
        frame : DynamicFrame
            The DynamicFrame to be written.
        connection_options : S3SinkConnectionOptions
            The connection options for S3 sink.
        format_options : XMLFormatOptions | None, optional
            The format options for XML. Defaults to None.
        transformation_ctx : str, optional
            The transformation context. Defaults to "".

        Returns
        -------
        DynamicFrame
            The DynamicFrame after writing to S3.

        """
        return self.write_dynamic_frame_from_options(
            frame=frame,
            connection_type=ConnectionType.S3.value,
            connection_options=connection_options,
            format=Format.XML.value,
            format_options=format_options,
            transformation_ctx=transformation_ctx,
        )


class S3Mixin(S3CSVMixin, S3ParquetMixin, S3JSONMixin, S3XMLMixin):
    """Mixin for working with S3 connections."""
